/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.osgicdi.impl;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.enterprise.context.spi.CreationalContext;
import javax.enterprise.inject.spi.Bean;
import javax.enterprise.inject.spi.BeanManager;

import org.glassfish.osgicdi.GeneralServiceEvent;
import org.glassfish.osgicdi.ServiceModified;
import org.glassfish.osgicdi.ServiceRegistered;
import org.glassfish.osgicdi.ServiceUnregistering;
import org.glassfish.osgijavaeebase.Extender;
import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleEvent;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceEvent;
import org.osgi.framework.ServiceListener;
import org.osgi.framework.ServiceReference;
import org.osgi.framework.ServiceRegistration;
import org.osgi.framework.SynchronousBundleListener;

/**
 * @author Tang Yong(tangyong@cn.fujitsu.com)
 */
public class OSGiCDIContainerActivator implements BundleActivator, ServiceListener, SynchronousBundleListener{
	
	private BundleContext context;
	private ServiceRegistration cdiExtenderReg;

	@Override
	public void start(BundleContext context) throws Exception {
		this.context = context;
		OSGiCDIExtender cdiExtender = new OSGiCDIExtender(context);
		cdiExtenderReg = context.registerService(Extender.class.getName(), cdiExtender, null);
		this.context.addServiceListener(this);
		context.addBundleListener(this);
	}

	@Override
	public void stop(BundleContext context) throws Exception {
		// Unregister so that ExtenderManager can stop Extender with an active bundle ctx
		cdiExtenderReg.unregister();		
	}

	@Override
	public void serviceChanged(ServiceEvent event) {
		ServiceReference serviceReference = event.getServiceReference();
		Object service = context.getService(serviceReference);
		if ((service instanceof RegisteredBeanManagerContext) || (service instanceof BeanManager)){
			return;
		}
		
		ServiceReference[] bmContextSRs = null;
		
		try {
			bmContextSRs = context.getServiceReferences(RegisteredBeanManagerContext.class.getName(), null);
		} catch (InvalidSyntaxException e) {
			e.printStackTrace();
		}
		
		if (bmContextSRs != null){
			GeneralServiceEvent cdiOSGiServiceEvent = null;
			switch(event.getType()){
			    case ServiceEvent.MODIFIED:
			    	cdiOSGiServiceEvent = new ServiceModified(serviceReference, context);
			    	
		        break;		            
		        case ServiceEvent.REGISTERED:
		        	cdiOSGiServiceEvent = new ServiceRegistered(serviceReference, context);
		        	
		        break;		        
		        case ServiceEvent.UNREGISTERING:
		        	cdiOSGiServiceEvent = new ServiceUnregistering(serviceReference, context);
		        	
		        break;
			}
			
			for (ServiceReference reference : bmContextSRs) {
				Object obj = context.getService(reference);
				RegisteredBeanManagerContext bmContext = null;
				
				if (obj != null && obj instanceof RegisteredBeanManagerContext){
					bmContext = (RegisteredBeanManagerContext)obj;
				}
				
				if (bmContext != null){
					BeanManager bm = bmContext.getBeanManager();
					ClassLoader tccl = bmContext.getTCCL();
					
					ClassLoader oldTCC = Thread.currentThread().getContextClassLoader(); 
					try{
						if (tccl != null){
							Thread.currentThread().setContextClassLoader(tccl);
						}
						
						if (cdiOSGiServiceEvent != null){
							fireCDIOSGiEvent(cdiOSGiServiceEvent, bm);
						}
						
					}catch(Throwable t){
						//t.printStackTrace();
					}finally{
						if (tccl != null){
							Thread.currentThread().setContextClassLoader(oldTCC);
						}			
					}					
				}			
			}
		}
	}

	private void fireCDIOSGiEvent(GeneralServiceEvent cdiOSGiServiceEvent,
			BeanManager bm) {
		List<Class<?>> classes = cdiOSGiServiceEvent.getServiceContractClasses();
		for (Class<?> clazz : classes) {
			Set<Annotation> set = new HashSet<Annotation>();
			set.add(new ServiceContractAnnotation(clazz));
			Annotation[] cdiOSGiEventQualifiers = set.toArray(new Annotation[set.size()]);
			bm.fireEvent(cdiOSGiServiceEvent, cdiOSGiEventQualifiers);
		}		
	}

	@Override
	public void bundleChanged(BundleEvent event) {
		 switch(event.getType()) {
	         case BundleEvent.STOPPING:
                  long bundleId = event.getBundle().getBundleId();
                  RegisteredBeanManagerContext bmContext = getBMCTXForCurrentBundle(bundleId);
                  
                  if (bmContext != null){
                	  BeanManager bm = bmContext.getBeanManager();
  					  ClassLoader tccl = bmContext.getTCCL();
  					
  					  ClassLoader oldTCC = Thread.currentThread().getContextClassLoader(); 
  					  try{
  						  if (tccl != null){
  							  Thread.currentThread().setContextClassLoader(tccl);
  						  }
  			
  						OSGiServiceExtension ext = getOSGiServiceExtension(bm);
  						List<ServiceRegistration> srs = ext.getBeanManagerContextSRS();
  						for(ServiceRegistration srg : srs){
  							srg.unregister();
  						}
  						srs.clear();
  					   }catch(Throwable t){
  						   //t.printStackTrace();
  					   }finally{
  						   if (tccl != null){
  							   Thread.currentThread().setContextClassLoader(oldTCC);
  						   }			
  					   }				
                     }              
             break;
		 }		
	}
	
	@SuppressWarnings("unchecked")
	private OSGiServiceExtension getOSGiServiceExtension(BeanManager bm){
		OSGiServiceExtension extension = null;
		Bean<OSGiServiceExtension> bean = (Bean<OSGiServiceExtension>) bm.resolve(bm.getBeans(OSGiServiceExtension.class));
        if (bean != null) {
            CreationalContext<OSGiServiceExtension> context = bm.createCreationalContext(bean);
            if (context != null) {
            	extension = (OSGiServiceExtension)bm.getReference(bean, OSGiServiceExtension.class, context);
            }
        }
		return extension;
	}

	private RegisteredBeanManagerContext getBMCTXForCurrentBundle(long bundleId) {
		RegisteredBeanManagerContext result = null;
		
        ServiceReference[] bmContextSRs = null;
		
		try {
			bmContextSRs = context.getServiceReferences(RegisteredBeanManagerContext.class.getName(), null);
			if (bmContextSRs != null){
				for (ServiceReference reference : bmContextSRs) {
					Object obj = context.getService(reference);
					RegisteredBeanManagerContext bmContext = null;
					
					if (obj != null && obj instanceof RegisteredBeanManagerContext){
						bmContext = (RegisteredBeanManagerContext)obj;
						
						long id = bmContext.getBundle().getBundleId();
						
						if ( bundleId != id ){
							continue;
						}
						
						result = bmContext;
						break;
					}
				}	
			}
		} catch (InvalidSyntaxException e) {
			e.printStackTrace();
		}
		
		return result;
	}
}
